/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corporation 2017. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package net.jazz.rtcext.scm.extensions.service;

import java.util.ArrayList;
import java.util.List;

import net.jazz.rtcext.scm.extensions.common.IScmExtensionsConstants;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.build.common.BuildItemFactory;
import com.ibm.team.build.common.model.IBuildDefinition;
import com.ibm.team.build.common.model.IBuildProperty;
import com.ibm.team.build.internal.common.ITeamBuildRequestService;
import com.ibm.team.build.internal.common.ITeamBuildService;
import com.ibm.team.build.internal.common.builddefinition.IJazzScmConfigurationElement;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IItemsResponse;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationParticipant;
import com.ibm.team.process.common.advice.runtime.IParticipantInfoCollector;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.scm.common.IScmService;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.IVersionableHandle;
import com.ibm.team.scm.common.IWorkspace;
import com.ibm.team.scm.common.dto.ICustomAttributeList;
import com.ibm.team.scm.common.process.ICommitItem;
import com.ibm.team.scm.common.process.ICommitOperationData;
import com.ibm.team.scm.common.process.ICommitOperationData.IResolvedPaths;

public class BuildOnCommitParticipant extends AbstractService implements IOperationParticipant {

	@Override
	public void run(AdvisableOperation operation, IProcessConfigurationElement element,
			IParticipantInfoCollector collector, IProgressMonitor monitor)
			throws TeamRepositoryException {
		
		// Get the commit operation data
		ICommitOperationData commitData = asCommitOperationData(operation.getOperationData());
		if (commitData == null) {
			// This is not the data we expected so return.
			return;
		}
		
		if (commitData.getWorkspace().isStream()) {
			// This can happen if changes are being made directly to a stream in the web UI.
			return;
		}
		
		CustomAttributeBasedConfiguration config = new CustomAttributeBasedConfiguration(commitData);
		
		// See if there is a build id associated with this component.
		String buildId = config.getBuildId();
		if (buildId == null) {
			// There was no build id specified for this component so just return.
			return;
		}
		
		// If there are changes in the commit that require a build, request a personal build.
		if (requiresBuild(config)) {
			requestPersonalBuild(commitData.getWorkspace(), buildId, collector);
		}
	}

	private ICommitOperationData asCommitOperationData(Object data) {
		if (data instanceof ICommitOperationData) {
			return (ICommitOperationData) data;
		}
		return null;
	}

	private boolean requiresBuild(CustomAttributeBasedConfiguration config) throws TeamRepositoryException {
		ICommitOperationData commitData = config.getCommitData();
		
		// We want to loop through all the commit items to see if there are any files
		List<IVersionableHandle> toCache = new ArrayList<>();
		
		// We want to loop through all the items that are being committed
		for (ICommitItem commitItem : commitData.getItems()) {
			// We are only interested in changes that match the change type specified.
			// We do this check first to reduce the amount of work we need to do fetching items
			// in order to determine the file name and path.
			if (config.changeTypeMatchesPattern(commitItem.getChangeType())) {
				// We are only interested in file changes
				IFileItemHandle fileHandle = getFileHandle(commitItem);
				if (fileHandle != null) {
					// We found a file so record the file so we can fetch the item to determine the name
					toCache.add(fileHandle);
				}
			}
		}
		
		// If there are no files changed, we don't need a build.
		if (toCache.isEmpty()) {
			return false;
		}
		
		// We need to prime the state cache will all the items we may need.
		// We do it this way so we can get them all in one fetch.
		commitData.cacheStates(toCache);
		
		// Loop through the files to see if any require a build.
		for (IVersionableHandle versionableHandle : toCache) {
			ICommitItem commitItem = commitData.getItem(versionableHandle.getItemId());
			if (fileChangeRequiresBuild(config, commitItem)) {
				return true;
			}
		}
		
		// Check the custom attributes on the files themselves.
		IScmService scmService = getService(IScmService.class);
		
		// We want to test the before states for the custom attributes because we are in the middle of a commit
		// and the custom attribute may not be available on the after state yet.
		IVersionableHandle[] versionablesArray = getBeforeStates(commitData.getItems());
		
		// Fetch the attributes for all our files together in a single query
		ICustomAttributeList[] attributesForVersionables = scmService.fetchCustomAttributesForVersionable(versionablesArray, null);
		
		// The input versionables array and output custom attributes array will be the same size and have corresponding entries
		for (int i = 0; i < attributesForVersionables.length; i++) {
			ICustomAttributeList attributeList = attributesForVersionables[i];
			IVersionableHandle versionableHandle = versionablesArray[i];
			
			// Check the attribute on each file to determine if a build is required
			Object buildWhenChanged = attributeList.getCustomAttributes().get(IScmExtensionsConstants.BUILD_WHEN_CHANGED_ATTRIBUTE);
			if (buildWhenChanged != null && buildWhenChanged.equals(Boolean.TRUE.toString())) {
				return true;
			}
		}
		
		return false;
	}

	private IVersionableHandle[] getBeforeStates(List<ICommitItem> items) {
		List<IVersionableHandle> result = new ArrayList<>(items.size());
		for (ICommitItem item : items) {
			if (item.getBeforeStateHandle() != null) {
				result.add(item.getBeforeStateHandle());
			}
		}
		return result.toArray(new IVersionableHandle[result.size()]);
	}

	/**
	 * Determine if a build is required for the given item.
	 * 
	 * @param commitData the commit data
	 * @param commitItem the commit item
	 * @param config the configuration from the custom attributes
	 * @return whether a build is require for the provided item
	 * @throws TeamRepositoryException
	 */
	private boolean fileChangeRequiresBuild(CustomAttributeBasedConfiguration config, ICommitItem commitItem) throws TeamRepositoryException {
		
		// If the name matches the name pattern, we will build
		String name = getFileName(config.getCommitData(), commitItem);
		if (config.fileNameMatchesPattern(name)) {
			return true;
		}
		
		// Or if the path matches the path pattern, we will build
		String path = getFilePath(config.getCommitData(), commitItem);
		if (config.filePathMatchesPattern(path)) {
			return true;
		}
		
		return false;
	}
	
	/**
	 * Return the after path for the item. In the case of a deletion, the before path is retuend.
	 * @param commitData the commit data
	 * @param commitItem the commit item
	 * @return the path of the item.
	 * @throws TeamRepositoryException
	 */
	private String getFilePath(ICommitOperationData commitData, ICommitItem commitItem) 
			throws TeamRepositoryException {
		
		// The commit data has a child object for resolving paths against the workspace.
		IResolvedPaths resolvedPaths = commitData.getResolvedPaths();
		if (commitItem.getAfterStateHandle()  == null) {
			// If we don't have an after state, we'll use the path for the before state
			return resolvedPaths.getBeforePath(commitItem);
		} else {
			// We have an after state so we'll use the path for that
			return resolvedPaths.getAfterPath(commitItem);
		}
	}

	private String getFileName(ICommitOperationData commitData, ICommitItem commitItem) throws TeamRepositoryException {
		IFileItemHandle fileHandle = getFileHandle(commitItem);
		IVersionable cachedState = commitData.getCachedState(fileHandle);
		return cachedState.getName();
	}

	private IFileItemHandle getFileHandle(ICommitItem commitItem) {	
		// The commit item has the before state, or the state that is already in the workspace,
		// and the after state, which is the state that will be in the workspace after the commit.
		if (commitItem.getAfterStateHandle()  == null) {
			// There is no after state so use the before state as the file handle, if it is a file
			if (commitItem.getBeforeStateHandle() instanceof IFileItemHandle) {
				return (IFileItemHandle) commitItem.getBeforeStateHandle();
			}
		} else if (commitItem.getAfterStateHandle() instanceof IFileItemHandle) {
			// We have an after state so return that as the state we care about
			return (IFileItemHandle) commitItem.getAfterStateHandle();
		}
		return null;
	}
	
	@SuppressWarnings("restriction")
	private void requestPersonalBuild(IWorkspace workspace, String buildId, IParticipantInfoCollector collector) throws TeamRepositoryException {
		
		ITeamBuildService buildService = getService(ITeamBuildService.class);
		IBuildDefinition buildDef = buildService.getBuildDefinition(buildId);
		
		/*
		 * If the build definition was found, try to run the build.
		 * Note: This code was borrowed from the RTC Extensions Workspace (https://jazz.net/library/article/1000)
		 */
		if (buildDef != null) {
			/*
			 * The team build request service is used to create, access and
			 * manipulate build requests. We will request a new personal build. The true
			 * and true in the build request parameters indicate that it is
			 * okay if multiple builds of this type are in the queue and that
			 * this is a personal build. The IBuildProperty parameter provides
			 * the workspace to use for the personal build.
			 * The  null means there are no build properties to
			 * delete.
			 */
			IBuildProperty userWorkspaceIdProperty = createBuildWorkspaceProperty(workspace);
			ITeamBuildRequestService brService = getService(ITeamBuildRequestService.class);
			IItemsResponse response = brService.requestBuild(buildDef, 
					new IBuildProperty[] { userWorkspaceIdProperty}, null,
					true, true);
			
			if ((response == null) || (response.getFirstClientItem() == null)) {
				
				/*
				 * The build request could not be created, report this back as
				 * an error. An error report will stop the work item save from
				 * succeeding and will show up in the team advisor.
				 */
				String description = NLS
						.bind(
								"The build request for build definition ''{0}'' could not be honored. It is likely that a build engine is not associated with the definition.",
								buildId);
				IReportInfo info = collector.createInfo(
						"Unable to request build", description);
				info.setSeverity(IProcessReport.ERROR);
				collector.addInfo(info);
			}
		} else {
			/*
			 * The build definition was not found, report this back as an error.
			 * An error report will stop the work item save from succeeding and
			 * will show up in the team advisor.
			 */
			String description = NLS
					.bind(
							"The build definition ''{0}'' could not be found.",
							buildId);
			IReportInfo info = collector.createInfo("Unable to request build",
					description);
			info.setSeverity(IProcessReport.ERROR);
			collector.addInfo(info);
		}
	}

	@SuppressWarnings("restriction")
	private IBuildProperty createBuildWorkspaceProperty(IWorkspace workspace) {
		IBuildProperty userWorkspaceIdProperty = BuildItemFactory.createBuildProperty();
		userWorkspaceIdProperty.setName(IJazzScmConfigurationElement.PROPERTY_WORKSPACE_UUID);
		userWorkspaceIdProperty.setValue(workspace.getItemId().getUuidValue());
		return userWorkspaceIdProperty;
	}

}
