/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corporation 2017. All Rights Reserved. 
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package net.jazz.rtcext.scm.extensions.service;

import java.util.Map;
import java.util.regex.Pattern;

import net.jazz.rtcext.scm.extensions.common.IScmExtensionsConstants;

import com.ibm.team.scm.common.process.ICommitItem;
import com.ibm.team.scm.common.process.ICommitOperationData;

/**
 * This class uses the custom attributes of a component to provide input into commit operation participants.
 */
public class CustomAttributeBasedConfiguration {
	
	private ICommitOperationData commitData;
	private boolean namePatternInitialized = false;
	private Pattern namePattern;
	private boolean pathPatternInitialized = false;
	private Pattern pathPattern;

	/**
	 * Create an instance for the given commit data
	 * @param commitData the commit data from the participant
	 */
	public CustomAttributeBasedConfiguration(ICommitOperationData commitData) {
		this.commitData = commitData;
	}

	/**
	 * Return the commit data
	 * @return the commit data
	 */
	public ICommitOperationData getCommitData() {
		return commitData;
	}
	
	/**
	 * Return the custom attribute of the component that matches the given key.
	 * The custom attribute needs to be defined in the project area that owns
	 * the component.
	 * 
	 * @param key the key
	 * @return the custom attribute of the component that matches the given key
	 */
	private String getComponentAttribute(String key) {
		// Get the custom attribute that identifies the build to run
		Map<String, Object> customAttributes = commitData.getComponent().getCustomAttributes();
		// Return the string associated with the key
		return (String)customAttributes.get(key);
	}
	
	/**
	 * Return the build id associated with the component or <code>null</code>
	 * if there is no build id associated.
	 * 
	 * @return the build id associated with the component or <code>null</code>
	 */
	public String getBuildId() {	
		return getComponentAttribute(IScmExtensionsConstants.BUILD_DEFINITION_ID_ATTRIBUTE);
	}
	
	/**
	 * If a file name matching expression has been specified for the component, use it to determine if
	 * a file with the given name should result in a build.
	 * 
	 * @param name the file name
	 * @return whether a change to the provided file should result in a build
	 */
	public boolean fileNameMatchesPattern(String name) {
		Pattern pattern = getNamePattern();
		if (pattern == null) {
			return false;
		}
		return pattern.matcher(name).matches();
	}
	
	/**
	 * If a file path matching expression has been specified for the component, use it to determine if
	 * a file with the given path should result in a build.
	 * 
	 * @param path the file path
	 * @return whether a change to the provided file path should result in a build
	 */
	public boolean filePathMatchesPattern(String path) {
		Pattern pattern = getPathPattern();
		if (pattern == null) {
			return false;
		}
		return pattern.matcher(path).matches();
	}
	
	public boolean changeTypeMatchesPattern(ICommitItem.ChangeType changeType) {
		String changeTypes = getComponentAttribute(IScmExtensionsConstants.CHANGE_TYPE_ATTRIBUTE);
		if (changeTypes == null) {
			// No type was specified so let's assume we want all change types
			return true;
		}
		// We are going to test for the most common change types
		if (changeTypes.equals("ALL")) {
			return true;
		} else if (changeType.equals(ICommitItem.ChangeType.ADD_TYPE)) {
			return changeTypes.contains("ADD");
		} else if (changeType.equals(ICommitItem.ChangeType.DELETE_TYPE)) {
			return changeTypes.contains("DELETE");
		} else if (changeType.equals(ICommitItem.ChangeType.MODIFY_TYPE)) {
			return changeTypes.contains("MODIFY");
		} else if (changeType.equals(ICommitItem.ChangeType.MOVE_RENAME_TYPE)) {
			return changeTypes.contains("MOVE");
		}
		return false;
	}
	
	private Pattern getNamePattern() {
		if (!namePatternInitialized) {
			namePatternInitialized = true;
			String regex = getComponentAttribute(IScmExtensionsConstants.FILE_NAME_PATTERN_ATTRIBUTE); // ".*\\.(c|h|cpp)";
			if (regex != null) {
				namePattern = Pattern.compile(regex);
			}
		}
		return namePattern;
	}
	
	private Pattern getPathPattern() {
		if (!pathPatternInitialized) {
			pathPatternInitialized = true;
			String regex = getComponentAttribute(IScmExtensionsConstants.FILE_PATH_PATTERN_ATTRIBUTE);
			if (regex != null) {
				pathPattern = Pattern.compile(regex);
			}
		}
		return pathPattern;
	}

}
